#include "lcipfgd.h"

BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
	return TRUE;
}

void IndexChannelConversationToOther(const char* SenderNick, const char* SenderUsername, const char* SenderAddress, const char* Command, const char* Target, const char* Trail, char* Output)
{
	string TaggedMessage = LimeChatColorsToHTML(Trail);
	string MessageLine = string(Target) + ": " + string(SenderNick) + ": " + TaggedMessage;
	double PresentTime = GetVariantTime();
	unsigned long ConversationID = GetConversationID(Target, PresentTime);

	CoInitialize(NULL);
	SendIndexingEvent(MessageLine.c_str(), Target, Target, NULL, ConversationID, PresentTime);
	CoUninitialize();
}

void IndexChannelConversationFromOther(const char* SenderNick, const char* SenderUsername, const char* SenderAddress, const char* Command, const char* Target, const char* Trail, char* Output)
{
	string TaggedMessage = mIRCColorsToHTML(Trail);
	string MessageLine = string(Target) + ": " + string(SenderNick) + ": " + TaggedMessage;
	double PresentTime = GetVariantTime();
	unsigned long ConversationID = GetConversationID(Target, PresentTime);

	CoInitialize(NULL);
	SendIndexingEvent(MessageLine.c_str(), Target, Target, NULL, ConversationID, PresentTime);
	CoUninitialize();
}

void IndexPrivateTalkToOther(const char* SenderNick, const char* SenderUsername, const char* SenderAddress, const char* Command, const char* Target, const char* Trail, char* Output)
{
	string TaggedMessage = LimeChatColorsToHTML(Trail);
	string MessageLine = string(SenderNick) + ": " + TaggedMessage;
	double PresentTime = GetVariantTime();
	unsigned long ConversationID = GetConversationID(Target, PresentTime);

	CoInitialize(NULL);
	SendIndexingEvent(MessageLine.c_str(), Target, Target, SenderNick, ConversationID, PresentTime);
	CoUninitialize();
}

void IndexPrivateTalkFromOther(const char* SenderNick, const char* SenderUsername, const char* SenderAddress, const char* Command, const char* Target, const char* Trail, char* Output)
{
	string TaggedMessage = mIRCColorsToHTML(Trail);
	string MessageLine = string(SenderNick) + ": " + TaggedMessage;
	double PresentTime = GetVariantTime();
	unsigned long ConversationID = GetConversationID(SenderNick, PresentTime);

	CoInitialize(NULL);
	SendIndexingEvent(MessageLine.c_str(), SenderNick, SenderNick, Target, ConversationID, PresentTime);
	CoUninitialize();
}

bool RegisterIndexingComponent()
{
	CComPtr<IGoogleDesktopRegistrar> spRegistrar;
	HRESULT hr = spRegistrar.CoCreateInstance(CLSID_GoogleDesktopRegistrar);
	if(FAILED(hr)) return false;

	CComSafeArray<VARIANT> descr(6);
	descr.SetAt(0, CComVariant(L"Title"));
	descr.SetAt(1, CComVariant(L"LimeChat Indexing Plugin for Google Desktop"));
	descr.SetAt(2, CComVariant(L"Description"));
	descr.SetAt(3, CComVariant(L"Indexes LimeChat Conversation"));
	descr.SetAt(4, CComVariant(L"Icon"));
	descr.SetAt(5, CComVariant(L""));
	hr = spRegistrar->StartComponentRegistration(CComBSTR(clsid), CComVariant(descr.m_psa));
	if(FAILED(hr)) return false;

	CComPtr<IGoogleDesktopRegisterIndexingPlugin> spRegistration;
	CComBSTR registrar_progid("GoogleDesktop.IndexingRegistration");
	hr = spRegistrar->GetRegistrationInterface(registrar_progid, reinterpret_cast<IUnknown**>(&spRegistration));
	if(SUCCEEDED(hr)) hr = spRegistration->RegisterIndexingPlugin(NULL);
	if(SUCCEEDED(hr)) hr = spRegistrar->FinishComponentRegistration();

	return SUCCEEDED(hr);
}

bool UnregisterIndexingComponent()
{
	CComPtr<IGoogleDesktopRegistrar> spRegistrar;
	HRESULT hr = spRegistrar.CoCreateInstance(CLSID_GoogleDesktopRegistrar);
	if(SUCCEEDED(hr)) hr = spRegistrar->UnregisterComponent(CComBSTR(clsid));

	return SUCCEEDED(hr);
}

bool SendIndexingEvent(const char* content, const char* buddy_name, const char* title, const char* user_name, const unsigned long conversation_id, const double message_time)
{
	CComPtr<IGoogleDesktopEventFactory> spFactory;
	HRESULT hr = spFactory.CoCreateInstance(CLSID_GoogleDesktop, NULL, CLSCTX_INPROC);
	if(FAILED(hr)) return false;

	CComPtr<IDispatch> spEventDisp;
	hr = spFactory->CreateEvent(CComBSTR(clsid), CComBSTR(L"Google.Desktop.IM"), &spEventDisp);
	if(hr == E_COMPONENT_NOT_REGISTERED){
		RegisterIndexingComponent();
		hr = spFactory->CreateEvent(CComBSTR(clsid), CComBSTR(L"Google.Desktop.IM"), &spEventDisp);
	}
	if(FAILED(hr)){
		UnregisterIndexingComponent();
		return false;
	}

	CComQIPtr<IGoogleDesktopEvent> spEvent(spEventDisp);
	if(spEvent == NULL) return false;

	if(SUCCEEDED(hr)) hr = spEvent->AddProperty(CComBSTR(L"message_time"), CComVariant(message_time, VT_DATE));
	if(SUCCEEDED(hr)) hr = spEvent->AddProperty(CComBSTR(L"format"), CComVariant("text/html"));
	if(SUCCEEDED(hr)) hr = spEvent->AddProperty(CComBSTR(L"content"), CComVariant(content));
	if(SUCCEEDED(hr)) hr = spEvent->AddProperty(CComBSTR(L"buddy_name"), CComVariant(buddy_name));
	if(SUCCEEDED(hr)) hr = spEvent->AddProperty(CComBSTR(L"title"), CComVariant(title));
	if(SUCCEEDED(hr) && conversation_id) hr = spEvent->AddProperty(CComBSTR(L"conversation_id"), CComVariant(conversation_id));
	if(SUCCEEDED(hr) && user_name) hr = spEvent->AddProperty(CComBSTR(L"user_name"), CComVariant(user_name));
	if(SUCCEEDED(hr)) hr = spEvent->Send(EventFlagIndexable);

	return SUCCEEDED(hr);
}

unsigned long GetConversationID(const string Identifier, const double PresentTime)
{
	CRegKey Registry;
	unsigned long ConversationID;
	double LastTime;
	long result;

	result = Registry.Create(HKEY_CURRENT_USER, "Software\\LimeChatIndexing", REG_NONE, REG_OPTION_VOLATILE);
	if(result != ERROR_SUCCESS) return 0;
	string EntryNameLastTime = Identifier + "_Time";
	string EntryNameConversationID = Identifier + "_ID";
	unsigned long size = sizeof(double);
	result = Registry.QueryBinaryValue(EntryNameLastTime.c_str(), &LastTime, &size);
	if(result == ERROR_SUCCESS && PresentTime - LastTime < 1.0 / 24 / 60 * CONVERSATION_TIMEOUT)
		result = Registry.QueryDWORDValue(EntryNameConversationID.c_str(), ConversationID);
	else result = ERROR_SUCCESS + 1;
	if(result != ERROR_SUCCESS){
		// Generate Conversation ID from PresentTime
		unsigned long *p = (unsigned long *)(&PresentTime);
		ConversationID = *p;
		Registry.SetDWORDValue(EntryNameConversationID.c_str(), ConversationID);
	}
	Registry.SetBinaryValue(EntryNameLastTime.c_str(), &PresentTime, sizeof(double));
	Registry.Close();

	return ConversationID;
}

double GetVariantTime()
{
	double VariantTime;
	SYSTEMTIME PresentTime;

	GetSystemTime(&PresentTime);
	SystemTimeToVariantTime(&PresentTime, &VariantTime);

	return VariantTime;
}

string mIRCColorsToHTML(string Message)
{
	size_t offset;
	
	// Escape Tag
	offset = 0;
	while((offset = Message.find('&', offset)) != string::npos) Message.replace(offset++, 1, "&amp;");
	offset = 0;
	while((offset = Message.find('<', offset)) != string::npos) Message.replace(offset, 1, "&lt;");
	offset = 0;
	while((offset = Message.find('>', offset)) != string::npos) Message.replace(offset, 1, "&gt;");
	offset = 0;
	while((offset = Message.find('"', offset)) != string::npos) Message.replace(offset, 1, "&quot;");

	// Bold Parse
	offset = 0;
	while((offset = Message.find('\x02', offset)) != string::npos){
        Message.replace(offset, 1, "<b>");
		offset = Message.find_first_of("\x02\x0F", offset);
		if(offset == string::npos){
			Message += "</b>";
			break;
		}
		else if(Message[offset] == '\x0F') Message.insert(offset, "</b>");
		else if(Message[offset] == '\x02') Message.replace(offset, 1, "</b>");
	}

	// Underline Parse
	offset = 0;
	while((offset = Message.find('\x1F', offset)) != string::npos){
        Message.replace(offset, 1, "<u>");
		offset = Message.find_first_of("\x1F\x0F", offset);
		if(offset == string::npos){
			Message += "</u>";
			break;
		}
		else if(Message[offset] == '\x0F') Message.insert(offset, "</u>");
		else if(Message[offset] == '\x1F') Message.replace(offset, 1, "</u>");
	}

	// Color, Reverse Parse
	bool color = false, bgcolor = false, reverse = false;
	int CharColor = 1, BGColor = 0;
	offset = 0;
	while((offset = Message.find_first_of("\x03\x16\x0F", offset)) != string::npos){
		if(color || reverse){
			string FontEndTag = "</font>";
			Message.insert(offset, FontEndTag);
			offset += FontEndTag.length();
		}
		// Stop
		if(Message[offset] == '\x0F'){
			Message.erase(offset, 1);
			color = false;
			bgcolor = false;
			reverse = false;
			CharColor = 1;
			BGColor = 0;
			continue;
		}
		// Reverse
		else if(Message[offset] == '\x16'){
			Message.erase(offset, 1);
			reverse = !reverse;
		}
		// Color
		else if(Message[offset] == '\x03'){
			Message.erase(offset, 1);
			if(!isdigit(Message[offset])){
				color = false;
				bgcolor = false;
				CharColor = 1;
				BGColor = 0;
			}
			else{
				color = true;
				if(isdigit(Message[offset+1])){
					CharColor = atoi(Message.substr(offset, 2).c_str());
					Message.erase(offset, 2);
				}
				else{
					CharColor = atoi(Message.substr(offset, 1).c_str());
					Message.erase(offset, 1);
				}
				if(Message[offset] == ','){
					if(isdigit(Message[offset+1])){
						bgcolor = true;
						if(isdigit(Message[offset+2])){
							BGColor = atoi(Message.substr(offset+1, 2).c_str());
							Message.erase(offset, 3);
						}
						else{
							BGColor = atoi(Message.substr(offset+1, 1).c_str());
							Message.erase(offset, 2);
						}
					}
				}
			}
		}
		if(color || reverse){
			string FontTag = "<font style=\"color:" + (reverse ? ReverseColorCodePage(CharColor) : ColorCodePage(CharColor));
			if(reverse || bgcolor) FontTag += ";background:" + (reverse ? ReverseColorCodePage(BGColor) : ColorCodePage(BGColor));
			FontTag += ";\">";
			Message.insert(offset, FontTag);
		}
	}
	if(color || reverse) Message += "</font>";

	return Message;
}

string LimeChatColorsToHTML(string Message)
{
	size_t offset;
	
	offset = 0;
	string BoldTag = "<bold>";
	while((offset = Message.find(BoldTag, offset)) != string::npos) Message.replace(offset, BoldTag.length(), "\x02");
	offset = 0;
	string UnderlineTag = "<underline>";
	while((offset = Message.find(UnderlineTag, offset)) != string::npos) Message.replace(offset, UnderlineTag.length(), "\x1F");
	offset = 0;
	string ReverseTag = "<reverse>";
	while((offset = Message.find(ReverseTag, offset)) != string::npos) Message.replace(offset, ReverseTag.length(), "\x16");
	offset = 0;
	string StopTag = "<stop>";
	while((offset = Message.find(StopTag, offset)) != string::npos) Message.replace(offset, StopTag.length(), "\x0F");

	offset = 0;
	string ColorTag = "<color";
	while((offset = Message.find(ColorTag, offset)) != string::npos){
		size_t tagend = Message.find('>', offset);
		if(tagend == string::npos) break;
		size_t space = offset + ColorTag.length();
		if(space == tagend){
			Message.replace(offset, tagend-offset+1, "\x03");
			continue;
		}
		if(Message[space] != ' '){
			offset = space;
			continue;
		}
		size_t comma = Message.find(',', space);
		if(comma == string::npos || comma > tagend){
			string CharColor = Message.substr(space+1, tagend-space-1);
			Message.replace(offset, tagend-offset+1, "\x03" + ColorNameToCode(CharColor));
		}
		else{
			string CharColor = Message.substr(space+1, comma-space-1);
			string BGColor = Message.substr(comma+1, tagend-comma-1);
			string CharColorCode = ColorNameToCode(CharColor);
			string BGColorCode = ColorNameToCode(BGColor);
			if(CharColorCode.empty() || BGColorCode.empty()) Message.replace(offset, tagend-offset+1, "\x03" + CharColorCode);
			else Message.replace(offset, tagend-offset+1, "\x03" + CharColorCode + "," + BGColorCode);
		}
	}

	return mIRCColorsToHTML(Message);
}
string ColorCodePage(int ColorCode)
{
	string CodePage[16] = {
		"white", "black", "navy", "green", "red", "maroon", "purple", "orange",
		"yellow", "lime", "teal", "aqua", "blue", "fuchsia", "gray", "silver"
	};

	return CodePage[ColorCode%16];
}

string ReverseColorCodePage(int ColorCode)
{
	string CodePage[16] = {
		"#000000", "#FFFFFF", "#FFFF7F", "#FF7FFF", "#00FFFF", "#7FFFFF", "#7FFF7F", "#007FFF",
		"#0000FF", "#FF00FF", "#FF7F7F", "#FF0000", "#FFFF00", "#00FF00", "#7F7F7F", "#3F3F3F"
	};

	return CodePage[ColorCode%16];
}

string ColorNameToCode(string ColorName)
{
	for(unsigned int i=0; i<ColorName.length(); i++) ColorName[i] = tolower(ColorName[i]);

	if(ColorName == "white")   return "00";
	if(ColorName == "black")   return "01";
	if(ColorName == "navy")    return "02";
	if(ColorName == "green")   return "03";
	if(ColorName == "red")     return "04";
	if(ColorName == "maroon")  return "05";
	if(ColorName == "purple")  return "06";
	if(ColorName == "orange")  return "07";
	if(ColorName == "yellow")  return "08";
	if(ColorName == "lime")    return "09";
	if(ColorName == "teal")    return "10";
	if(ColorName == "aqua")    return "11";
	if(ColorName == "blue")    return "12";
	if(ColorName == "fuchsia") return "13";
	if(ColorName == "gray")    return "14";
	if(ColorName == "silver")  return "15";
	else                       return "";
}
